import re
from typing import Dict, Match, Pattern

"""
Copyright 2017 Cornell University

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

"""Convert between TeX escapes and UTF8."""

# fmt: off
accents = {
    # first accents with non-letter prefix, e.g. \'A
    "'A": 0x00c1, "'C": 0x0106, "'E": 0x00c9, "'I": 0x00cd,
    "'L": 0x0139, "'N": 0x0143, "'O": 0x00d3, "'R": 0x0154,
    "'S": 0x015a, "'U": 0x00da, "'Y": 0x00dd, "'Z": 0x0179,
    "'a": 0x00e1, "'c": 0x0107, "'e": 0x00e9, "'i": 0x00ed,
    "'l": 0x013a, "'n": 0x0144, "'o": 0x00f3, "'r": 0x0155,
    "'s": 0x015b, "'u": 0x00fa, "'y": 0x00fd, "'z": 0x017a,
    '"A': 0x00c4, '"E': 0x00cb, '"I': 0x00cf, '"O': 0x00d6,
    '"U': 0x00dc, '"Y': 0x0178, '"a': 0x00e4, '"e': 0x00eb,
    '"i': 0x00ef, '"o': 0x00f6, '"u': 0x00fc, '"y': 0x00ff,
    '.A': 0x0226, '.C': 0x010a, '.E': 0x0116, '.G': 0x0120,
    '.I': 0x0130, '.O': 0x022e, '.Z': 0x017b, '.a': 0x0227,
    '.c': 0x010b, '.e': 0x0117, '.g': 0x0121, '.o': 0x022f,
    '.z': 0x017c, '=A': 0x0100, '=E': 0x0112, '=I': 0x012a,
    '=O': 0x014c, '=U': 0x016a, '=Y': 0x0232, '=a': 0x0101,
    '=e': 0x0113, '=i': 0x012b, '=o': 0x014d, '=u': 0x016b,
    '=y': 0x0233, '^A': 0x00c2, '^C': 0x0108, '^E': 0x00ca,
    '^G': 0x011c, '^H': 0x0124, '^I': 0x00ce, '^J': 0x0134,
    '^O': 0x00d4, '^S': 0x015c, '^U': 0x00db, '^W': 0x0174,
    '^Y': 0x0176, '^a': 0x00e2, '^c': 0x0109, '^e': 0x00ea,
    '^g': 0x011d, '^h': 0x0125, '^i': 0x00ee, '^j': 0x0135,
    '^o': 0x00f4, '^s': 0x015d, '^u': 0x00fb, '^w': 0x0175,
    '^y': 0x0177, '`A': 0x00c0, '`E': 0x00c8, '`I': 0x00cc,
    '`O': 0x00d2, '`U': 0x00d9, '`a': 0x00e0, '`e': 0x00e8,
    '`i': 0x00ec, '`o': 0x00f2, '`u': 0x00f9, '~A': 0x00c3,
    '~I': 0x0128, '~N': 0x00d1, '~O': 0x00d5, '~U': 0x0168,
    '~a': 0x00e3, '~i': 0x0129, '~n': 0x00f1, '~o': 0x00f5,
    '~u': 0x0169,
    # and now ones with letter prefix \c{c} etc..
    'HO': 0x0150, 'HU': 0x0170, 'Ho': 0x0151, 'Hu': 0x0171,
    'cC': 0x00c7, 'cE': 0x0228,
    'cG': 0x0122, 'cK': 0x0136, 'cL': 0x013b, 'cN': 0x0145,
    'cR': 0x0156, 'cS': 0x015e, 'cT': 0x0162, 'cc': 0x00e7,
    'ce': 0x0229, 'cg': 0x0123, 'ck': 0x0137, 'cl': 0x013c,
    # Commented out due ARXIVDEV-2322 (bug reported by PG)
    # 'ci' : 'i\x{0327}' = chr(0x69).ch(0x327) # i with combining cedilla
    'cn': 0x0146, 'cr': 0x0157, 'cs': 0x015f, 'ct': 0x0163,
    'kA': 0x0104, 'kE': 0x0118, 'kI': 0x012e, 'kO': 0x01ea,
    'kU': 0x0172, 'ka': 0x0105, 'ke': 0x0119, 'ki': 0x012f,
    'ko': 0x01eb, 'ku': 0x0173, 'rA': 0x00c5, 'rU': 0x016e,
    'ra': 0x00e5, 'ru': 0x016f, 'uA': 0x0102, 'uE': 0x0114,
    'uG': 0x011e, 'uI': 0x012c, 'uO': 0x014e, 'uU': 0x016c,
    'ua': 0x0103, 'ue': 0x0115, 'ug': 0x011f,
    'ui': 0x012d, 'uo': 0x014f, 'uu': 0x016d,
    'vA': 0x01cd, 'vC': 0x010c, 'vD': 0x010e,
    'vE': 0x011a, 'vG': 0x01e6, 'vH': 0x021e, 'vI': 0x01cf,
    'vK': 0x01e8, 'vL': 0x013d, 'vN': 0x0147, 'vO': 0x01d1,
    'vR': 0x0158, 'vS': 0x0160, 'vT': 0x0164, 'vU': 0x01d3,
    'vZ': 0x017d, 'va': 0x01ce, 'vc': 0x010d, 'vd': 0x010f,
    've': 0x011b, 'vg': 0x01e7, 'vh': 0x021f, 'vi': 0x01d0,
    'vk': 0x01e9, 'vl': 0x013e, 'vn': 0x0148, 'vo': 0x01d2,
    'vr': 0x0159, 'vs': 0x0161, 'vt': 0x0165, 'vu': 0x01d4,
    'vz': 0x017e
}
r"""
Hash to lookup tex markup and convert to Unicode.

macron: a line above character (overbar \={} in TeX)
caron: v-shape above character (\v{ } in TeX)
See: http://www.unicode.org/charts/

"""

textlet = {
    'AA': 0x00c5, 'AE': 0x00c6, 'DH': 0x00d0, 'DJ': 0x0110,
    'ETH': 0x00d0, 'L': 0x0141, 'NG': 0x014a, 'O': 0x00d8,
    'oe': 0x0153, 'OE': 0x0152, 'TH': 0x00de, 'aa': 0x00e5,
    'ae': 0x00e6,
    'dh': 0x00f0, 'dj': 0x0111, 'eth': 0x00f0, 'i': 0x0131,
    'l': 0x0142, 'ng': 0x014b, 'o': 0x00f8, 'ss': 0x00df,
    'th': 0x00fe,
    # Greek (upper)
    'Gamma': 0x0393, 'Delta': 0x0394, 'Theta': 0x0398,
    'Lambda': 0x039b, 'Xi': 0x039E, 'Pi': 0x03a0,
    'Sigma': 0x03a3, 'Upsilon': 0x03a5, 'Phi': 0x03a6,
    'Psi': 0x03a8, 'Omega': 0x03a9,
    # Greek (lower)
    'alpha': 0x03b1, 'beta': 0x03b2, 'gamma': 0x03b3,
    'delta': 0x03b4, 'epsilon': 0x03b5, 'zeta': 0x03b6,
    'eta': 0x03b7, 'theta': 0x03b8, 'iota': 0x03b9,
    'kappa': 0x03ba, 'lambda': 0x03bb, 'mu': 0x03bc,
    'nu': 0x03bd, 'xi': 0x03be, 'omicron': 0x03bf,
    'pi': 0x03c0, 'rho': 0x03c1, 'varsigma': 0x03c2,
    'sigma': 0x03c3, 'tau': 0x03c4, 'upsion': 0x03c5,
    'varphi': 0x03C6,  # φ
    'phi':  0x03D5,  # ϕ
    'chi': 0x03c7, 'psi': 0x03c8, 'omega': 0x03c9,
}


def _p_to_match(tex_to_chr: Dict[str, int]) -> Pattern:
    # textsym and textlet both use the same sort of regex pattern.
    keys = r'\\(' + '|'.join(tex_to_chr.keys()) + ')'
    pstr = r'({)?' + keys + r'(\b|(?=_))(?(1)}|(\\(?= )| |{}|)?)'
    return re.compile(pstr)


textlet_pattern = _p_to_match(textlet)

textsym = {
    'P': 0x00b6, 'S': 0x00a7, 'copyright': 0x00a9,
    'guillemotleft': 0x00ab, 'guillemotright': 0x00bb,
    'pounds': 0x00a3, 'dag': 0x2020, 'ddag': 0x2021,
    'div': 0x00f7, 'deg': 0x00b0}

# fmt: on

textsym_pattern = _p_to_match(textsym)


def _textlet_sub(match: Match) -> str:
    return chr(textlet[match.group(2)])


def _textsym_sub(match: Match) -> str:
    return chr(textsym[match.group(2)])


def texch2UTF(acc: str) -> str:
    """Convert single character TeX accents to UTF-8.

    Strip non-whitepsace characters from any sequence not recognized (hence
    could return an empty string if there are no word characters in the input
    string).

    chr(num) will automatically create a UTF8 string for big num
    """
    if acc in accents:
        return chr(accents[acc])
    else:
        return re.sub(r"[^\w]+", "", acc, flags=re.IGNORECASE)


def tex2utf(tex: str, letters: bool = True) -> str:
    r"""Convert some TeX accents and greek symbols to UTF-8 characters.

    :param tex: Text to filter.

    :param letters: If False, do not convert greek letters or
    ligatures.  Greek symbols can cause problems. Ex. \phi is not
    suppose to look like φ. φ looks like \varphi.  See ARXIVNG-1612

    :returns: string, possibly with some TeX replaced with UTF8

    """
    # Do dotless i,j -> plain i,j where they are part of an accented i or j
    utf = re.sub(r"/(\\['`\^\"\~\=\.uvH])\{\\([ij])\}", r"\g<1>\{\g<2>\}", tex)

    # Now work on the Tex sequences, first those with letters only match
    if letters:
        utf = textlet_pattern.sub(_textlet_sub, utf)

    utf = textsym_pattern.sub(_textsym_sub, utf)

    utf = re.sub(r"\{\\j\}|\\j\s", "j", utf)  # not in Unicode?

    # reduce {{x}}, {{{x}}}, ... down to {x}
    while re.search(r"\{\{([^\}]*)\}\}", utf):
        utf = re.sub(r"\{\{([^\}]*)\}\}", r"{\g<1>}", utf)

    # Accents which have a non-letter prefix in TeX, first \'e
    utf = re.sub(r'\\([\'`^"~=.][a-zA-Z])', lambda m: texch2UTF(m.group(1)), utf)

    # then \'{e} form:
    utf = re.sub(
        r'\\([\'`^"~=.])\{([a-zA-Z])\}',
        lambda m: texch2UTF(m.group(1) + m.group(2)),
        utf,
    )

    # Accents which have a letter prefix in TeX
    #  \u{x} u above (breve), \v{x}   v above (caron), \H{x}   double accute...
    utf = re.sub(
        r"\\([Hckoruv])\{([a-zA-Z])\}",
        lambda m: texch2UTF(m.group(1) + m.group(2)),
        utf,
    )

    # Don't do \t{oo} yet,
    utf = re.sub(r"\\t{([^\}])\}", r"\g<1>", utf)

    # bdc34: commented out in original Perl
    # $utf =~ s/\{(.)\}/$1/g; #  remove { } from around {x}

    return utf



"""Package constants and settings."""

import logging
import logging.config

import colorful as cf
from platformdirs import user_data_path, user_log_path

# set globals
VOY_PATH = user_data_path("voy", ensure_exists=True)
VOY_LOGS = user_log_path("voy", ensure_exists=True) / "voy.log"
CATEGORIES = ["cs.CV", "cs.LG", "cs.CL", "cs.AI", "cs.NE", "cs.RO"]


# set logging
logging.config.dictConfig(
    config={
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "simple": {"format": "%(levelname)s: %(message)s"},
            "detailed": {
                "format": "[%(asctime)s %(levelname)s %(module)s:L%(lineno)d] %(message)s",
                "datefmt": "%Y-%m-%dT%H:%M:%S%z",
            },
        },
        "handlers": {
            "stderr": {
                "class": "logging.StreamHandler",
                "formatter": "simple",
                "level": "CRITICAL",
                "stream": "ext://sys.stderr",
            },
            "stdout": {
                "class": "logging.StreamHandler",
                "formatter": "simple",
                "level": "WARNING",
                "stream": "ext://sys.stdout",
            },  # configured but unused
            "file": {
                "backupCount": 5,
                "class": "logging.handlers.RotatingFileHandler",
                "filename": VOY_LOGS,
                "formatter": "detailed",
                "level": "DEBUG",
                "maxBytes": 100_000,
            },
        },
        "loggers": {"root": {"level": "DEBUG", "handlers": ["stderr", "file"]}},
    }
)

# get application logger
log = logging.getLogger("voy")


# set colors
cf.use_8_ansi_colors()


log.info("set up complete.")



"""SQL queries are defined here until we figure them out."""

create_tables = [
    """--sql
    -- author
    CREATE TABLE IF NOT EXISTS author (
        id TEXT PRIMARY KEY,
        last_name TEXT NOT NULL,
        other_names TEXT NOT NULL,
        name_suffix TEXT,
        followed INTEGER CHECK( followed in (0, 1) ) NOT NULL
    ) STRICT
    """,
    """--sql
    -- paper
    CREATE TABLE IF NOT EXISTS paper (
        id TEXT PRIMARY KEY,
        updated TEXT NOT NULL,
        created TEXT NOT NULL,
        meta TEXT NOT NULL,
        visible INTEGER CHECK( visible in (0, 1) ) DEFAULT 1 NOT NULL
    ) STRICT
    """,
    """--sql
    --support for full text search
    CREATE VIRTUAL TABLE IF NOT EXISTS paper_fts USING FTS5(
        id,
        updated,
        created,
        meta,
        visible,
        content='paper',
        content_rowid='rowid'
    )
    """,
    """--sql
    --TODO: I think we should avoid executing this every time.
    INSERT INTO paper_fts (rowid, id, updated, created, meta, visible)
    SELECT rowid, id, updated, created, meta, visible FROM paper
    """,
    """--sql
    -- authorship
    CREATE TABLE IF NOT EXISTS authorship (
        id INTEGER PRIMARY KEY,
        author_id TEXT NOT NULL,
        paper_id TEXT NOT NULL,
        FOREIGN KEY (author_id) REFERENCES author (id),
        FOREIGN KEY (paper_id) REFERENCES paper (id)
    ) STRICT
    """,
    """--sql
    -- create indices
    CREATE INDEX IF NOT EXISTS name_idx ON author(last_name, other_names)
    """,
    """--sql
    CREATE INDEX IF NOT EXISTS followed_idx ON author(followed)
    """,
    """--sql
    CREATE INDEX IF NOT EXISTS author_idx ON authorship(author_id)
    """,
]


# author  -----------------------------------------------------------------------------

is_author = """--sql
    SELECT EXISTS(SELECT 1 FROM author WHERE author.id = :id)
"""

add_author = """--sql
    INSERT INTO author (id, last_name, other_names, name_suffix, followed)
    VALUES (:id, :last_name, :other_names, :name_suffix, :followed);
"""

get_author = """--sql
    SELECT * FROM author WHERE author.id = :id
"""

get_followed = """--sql
    SELECT * FROM author WHERE author.followed = 1
"""

search_author = """--sql
    SELECT *
      FROM author
     WHERE last_name = :last_name
       AND other_names = :other_names
"""

search_by_last_name = """--sql
    SELECT *
      FROM author
     WHERE last_name = :last_name
"""

count_author = """--sql
    SELECT COUNT(*) FROM author
"""

follow_author = """--sql
    UPDATE author SET followed = :followed WHERE id = :id
"""

is_followed = """--sql
    SELECT author.followed FROM author WHERE author.id = :id
"""

count_followed = """--sql
    SELECT COUNT(*) FROM author WHERE author.followed = 1
"""

# paper  ------------------------------------------------------------------------------

is_paper = """--sql
    SELECT EXISTS(SELECT 1 FROM paper WHERE paper.id = :id)
"""

add_paper = """--sql
    INSERT INTO paper (id, updated, created, meta)
    VALUES (:id, :updated, :created, :meta);
"""

update_paper = """--sql
    UPDATE paper
        SET created = :created,
            updated = :updated,
            meta = :meta,
            visible = :visible
        WHERE id = :id
"""

get_paper = """--sql
    SELECT *
      FROM paper
     WHERE paper.id = :id
"""

fts = """--sql
    SELECT *
      FROM paper_fts
     WHERE paper_fts
     MATCH :term
"""

count_paper = """--sql
    SELECT COUNT(*) FROM paper
"""

last_paper_by = """--sql
    SELECT * FROM paper ORDER BY {col:s} DESC LIMIT 1
"""

# authorship --------------------------------------------------------------------------

add_authorship = """--sql
    INSERT INTO authorship (author_id, paper_id)
    VALUES (:author_id, :paper_id);
"""


get_papers_by_author_id = """--sql
SELECT paper.* FROM
(SELECT * FROM author WHERE author.id = :aid) AS a
INNER JOIN authorship
    ON a.id = authorship.author_id
INNER JOIN paper
    ON authorship.paper_id = paper.id
WHERE paper.updated BETWEEN :starting_date AND datetime("now")
"""


get_papers_by_author_name = """--sql
SELECT paper.* FROM
(SELECT * FROM author
    WHERE author.last_name = :last_name
        AND author.other_names = :other_names
) AS a
INNER JOIN authorship
    ON a.id = authorship.author_id
INNER JOIN paper
    ON authorship.paper_id = paper.id
WHERE paper.updated BETWEEN :starting_date AND datetime("now")
"""


get_coauthors = """--sql
SELECT author.* FROM (SELECT * FROM paper WHERE paper.id = :pid) AS p
INNER JOIN authorship
    ON authorship.paper_id = p.id
INNER JOIN author
    ON author.id = authorship.author_id;
"""



from __future__ import annotations

import curses
import logging
import math
import os
import re
from collections import defaultdict
from datetime import datetime as dt
from itertools import chain

from . import VOY_LOGS, cf
from .models import Author, Paper

log = logging.getLogger("voy")


try:
    WIDTH = os.get_terminal_size().columns
except OSError:
    WIDTH = 80


def short_date(date):
    t = dt.strptime(date, Paper.DATE_FMT)
    return dt.strftime(t, "%Y %b %d")


def _clip(seq: str, offset: int = 0) -> str:
    if len(seq) + offset > WIDTH:
        return f"{seq:.{WIDTH - (offset + 3)}s}..."
    return seq


def _date_title_rows(
    papers: list[Paper], offset: int = 0
) -> tuple[tuple[str, str, str], ...]:
    """Assumes an ordered set and returns tuples of sparse (year, month_day,
    title).  Year and month are sparse, so that we don't display redundant info,
    titles are clipped to the available row width.

    Dec 01 title..
        24 title..
    2024 ─────────
    Jan 20 title..
        18 title..
        12 title..
    Feb 23 title..
    ...
    """
    rows = []
    _year, _month = "", ""
    for paper in papers:
        year, month, day = short_date(paper.updated).split(" ")
        if _month != month:
            _month = month
            m = month
        else:
            m = " " * len(month)
        y = ""
        if _year != year:
            if _year:
                _year = year
                y = year
            _year = year
        month_day = f"{m} {day}"
        year_head = f"{y} {'\u2500' * (WIDTH - len(month_day))}\n" if y else ""
        clipped_title = _clip(paper.meta.title, len(month_day) + offset)
        rows.append((year_head, month_day, clipped_title))
    return tuple(rows)


def _coauthor_rows(paper, indent_len):
    cos = []
    for author in paper.meta.authors:
        # TODO fix this type instability by moving out authors from meta?
        last, other, sfx = list(author.values())[:3]
        name = f"{other} {last} {sfx}" if sfx else f"{other} {last}"
        cos.append(name)
    w = os.get_terminal_size().columns - indent_len
    line = ", ".join(cos)
    if len(line) < w:
        return f"{' ' * indent_len}{line}"
    co_str = cos.pop()
    # make the rows
    for co in cos:
        lines_so_far = co_str.split("\n")
        if len(lines_so_far[-1] + co) + 2 < w:
            co_str += ", " + co
        else:
            co_str += f",\n{' ' * indent_len}{cos.pop()}"
            # some papers just have sooooo many authors
            if len(lines_so_far) == 2 and len(cos) > 2:
                # assumes there's space for three authors per line
                # plus the extra markup
                co_str += f", ..., ..., {cos[-2]}, {cos[-1]}"
                break
    return f"{' ' * indent_len}{co_str}"


def _list_papers(papers: list[Paper], coauthors: bool, url: bool, pfx=" ", sep=" "):
    """Lists paper title, date, co-authors and url, in the formats:

    {pfx}MM dd{sep}Title...

    {pfx}MM dd{sep}Title...
                   co-authors

    {pfx}MM dd{sep}Title...
                   url

    {pfx}MM dd{sep}Title...
                   co-authors
                   url
    """
    offset = len(pfx) + len(sep)
    for (year, date, title), paper in zip(_date_title_rows(papers, offset), papers):
        indent = len(date) + offset
        c_year = cf.dimmed | year
        c_date = cf.bold | date
        c_title = cf.bold | title if coauthors else title
        print(f"{c_year}{pfx}{c_date}{sep}{c_title}")
        if coauthors:
            print(_coauthor_rows(paper, indent))
        if url:
            print(cf.cyan | f"{'':>{indent}}https://arxiv.org/abs/{paper.id}")


def latest_papers(data: list[Author], num_papers: int, coauthors: bool, url: bool):
    slice_ = slice(0, num_papers or None)
    papers = sorted(
        set(chain.from_iterable([author.papers for author in data])),
        key=lambda p: p.updated,
        reverse=True,
    )[slice_]

    _list_papers(papers, coauthors, url, pfx="")


def paper_list(data: list[Paper], coauthors: bool, url: bool):
    """View of the paper titles."""
    papers = sorted(data, key=lambda p: p.updated, reverse=True)
    _list_papers(papers, coauthors=coauthors, url=url, pfx="")


def author_paper_list(author: Author, num_papers: int, coauthors: bool, url: bool):
    slice_ = slice(0, num_papers or None)
    papers = sorted(author.papers, key=lambda p: p.updated, reverse=True)[slice_]

    # author header
    print(cf.yellow | author, "({})".format(cf.green | f"{len(author.papers)} papers"))

    _list_papers(papers, coauthors, url)


def author_list(authors):
    max_name_len = max([len(str(a)) for a in authors])
    for author in authors:
        print(
            "{0:{2}s}{1} |  {3}".format(
                str(author),
                cf.green | "followed" if author.followed else "",
                max_name_len + 3,
                cf.yellow | author.id,
            )
        )


def _list_on_cols(data: list[str], cols: int = 1, sep_size: int = 3) -> None:
    rows = math.ceil(len(data) / cols)

    lines: dict[int, list] = defaultdict(list)
    max_len: dict[int, int] = {}
    for count, item in enumerate(data):
        lines[count % rows].append(item)
        max_len[count // rows] = max(max_len.get(count // rows, 0), len(item))

    for _, line in sorted(lines.items()):
        for cidx, item in enumerate(line):
            print(item.ljust(max_len[cidx] + sep_size), end="")
        print()


def author_table(authors: list[Author]) -> None:
    if authors[0].papers is not None:
        data = [
            "{} ({})".format(author, len(author.papers)) for author in sorted(authors)
        ]
    else:
        data = [str(author) for author in sorted(authors)]
    author_lenghts = sorted([len(s) for s in data], reverse=True)
    cols, sep = 5, 3
    while (sum(author_lenghts[:cols]) + (sep * cols)) > WIDTH:
        cols -= 1

    if authors[0].papers is not None:
        data = [
            "{} ({})".format(author, cf.green | len(author.papers))
            for author in sorted(authors)
        ]

    _list_on_cols(data, cols=cols)


def wordlist(string):
    """Split the string into individual elements and returning them as a list."""
    return re.split(r"(\s+|\n+)", string)


def wordwrap(window: curses._CursesWindow, string: str) -> None:
    """Word wrapper for Python Curses module by Ikaros Ainasoja."""

    # Get cursor position
    cursor_y, cursor_x = window.getyx()

    # Get window dimensions
    win_height, win_width = window.getmaxyx()

    # If string length <= window width: print the string.
    if len(string) + cursor_x <= win_width:
        window.addstr(string)

    # Otherwise, split it into individual words and whitespaces,
    # put them in a list and try to print them one at a time.
    else:
        for item in wordlist(string):
            # Skip spaces in the beginning of a new line.
            if cursor_x == 0 and item == " ":
                continue

            # If list item lenght <= distance to window edge: print it.
            if len(item) + cursor_x <= win_width:
                # we are on the last row, with space for just one character left.
                # the item fits (len=1) but it pushes the cursor out of bounds.
                if len(item) + cursor_x == win_width and cursor_y == win_height - 1:
                    return
                window.addstr(item)

            # Otherwise, move to the next line and try to fit it there.
            else:
                # If this would move the cursor out of bounds: error.
                if cursor_y == win_height - 1:
                    return

                # Otherwise, print it.
                window.addstr(cursor_y + 1, 0, item)

            # Get cursor position before the next list item.
            cursor_y, cursor_x = window.getyx()


def init_swipe_view(stdscr):
    # hide cursor
    curses.curs_set(0)
    # start colors in curses
    curses.start_color()
    curses.use_default_colors()
    curses.init_pair(1, curses.COLOR_CYAN, -1)
    curses.init_pair(2, curses.COLOR_YELLOW, -1)

    def f(paper, status) -> None:
        # initialization
        stdscr.clear()
        stdscr.refresh()

        # terminal size
        height, width = stdscr.getmaxyx()

        # declare strings
        title: str = paper.meta.title
        abstract: str = paper.meta.abstract.replace("\n", " ")
        url = f"https://arxiv.org/abs/{paper.id}"
        footer = f"{status} | \u2190 (junk), \u2192 (keep), \u2193 (back), 'q' (exit)."

        # padding
        pad = 0.05
        start_x = int(width * pad)
        start_y = int(height * pad)
        content_width = int(width * (1 - pad * 2))
        title_rows = 1 if len(title) < content_width else 2
        url_rows, ftr_rows = 1, 1
        abs_rows = height - start_y - (title_rows + url_rows + ftr_rows) - 1
        abs_start_y = start_y + title_rows + url_rows + 1

        # title
        title_win = curses.newwin(title_rows, content_width, start_y, start_x)
        title_win.clear()
        title_win.attrset(curses.A_BOLD if paper.visible else curses.A_NORMAL)
        wordwrap(title_win, title)

        # url
        url_win = curses.newwin(url_rows, content_width, start_y + title_rows, start_x)
        url_win.clear()
        url_win.attrset(curses.color_pair(1) if paper.visible else curses.A_DIM)
        url_win.addstr(url)

        # abstract
        abstract_win = curses.newwin(abs_rows, content_width, abs_start_y, start_x)
        abstract_win.clear()
        abstract_win.attrset(curses.A_NORMAL if paper.visible else curses.A_DIM)
        wordwrap(abstract_win, abstract)

        # status bar
        footer_win = curses.newwin(ftr_rows, content_width, height - 1, start_x)
        footer_win.clear()
        footer_win.addstr(footer)

        # refresh everything
        url_win.refresh()
        title_win.refresh()
        abstract_win.refresh()
        footer_win.refresh()

    return f


# logging views ------------------------------------------------------------------------


def info(msg):
    log.info(msg)
    print(msg)


def critical(msg, *args, **kwargs):
    """Critical messages are the only ones communicated to the user."""
    msg = str(msg) + f"Check logs for details\n{VOY_LOGS}."
    log.critical(msg, *args, **kwargs)



"""This file plays two roles:
- contains all the Controllers that orchestrate user inputs, models and views.
- defines the argparser.
- is the entry point of the program :)
"""

from __future__ import annotations

import argparse
import csv
import curses
import logging
from collections.abc import Sequence
from dataclasses import MISSING
from datetime import datetime as dt
from itertools import chain
from pathlib import Path

from arxiv import UnexpectedEmptyPageError
from datargs import arg, argsclass, parse

from . import VOY_LOGS, VOY_PATH
from . import query as Q
from . import views as V
from .models import Author, AuthorArxiv, AuthorDB, Paper, PaperDB
from .seed import from_json
from .storage import Storage

log = logging.getLogger("voy")


# TODO: exit codes!!!


def show(opt: Show) -> None:
    with Storage() as db:
        if opt.author:  # fetch for one author
            authors = AuthorDB(db).search(" ".join(opt.author))
            if not authors:
                V.info(f"Found no author {opt.author}")
                return
            for author in authors:
                AuthorDB(db).get_papers_(author, opt.since, True)
        else:  # fetch for all followees
            authors = AuthorDB(db).get_followees()
            for followee in authors:
                AuthorDB(db).get_papers_(followee, opt.since, True)

    # list papers
    sorted_authors = sorted(authors, key=lambda author: author.other_names)
    if opt.by_author or opt.author:
        for author in sorted_authors:
            V.author_paper_list(author, opt.num, opt.coauthors, opt.url)
    else:
        V.latest_papers(sorted_authors, opt.num, opt.coauthors, opt.url)


def search_author_in_db(searched: Sequence[str]) -> None:
    with Storage() as db:
        authors = AuthorDB(db).search(" ".join(searched))
    V.author_list(authors)


def search_text_in_db(searched: Sequence[str], coauthors: bool, url: bool) -> None:
    with Storage() as db:
        papers = PaperDB(db).full_text_search(searched)
    V.paper_list(papers, coauthors, url)


def search_author_in_arxiv(searched: Sequence[str], num_papers: int) -> None:
    res = AuthorArxiv.search(" ".join(searched), 100)
    authors = sorted(res, key=lambda a: (a.other_names, a.last_name))
    for author in authors:
        V.author_paper_list(author, num_papers, False, False)


def _one_year_back() -> str:
    now = dt.now()
    try:
        since = now.replace(year=now.year - 1)
    except ValueError:
        since = now.replace(year=now.year - 1, day=now.day - 1)
    return dt.strftime(since, Paper.DATE_FMT)


def triage(opt: Triage) -> None:
    num_papers = 0

    with Storage() as db:
        authors = AuthorDB(db).get_followees()
        for followee in authors:
            AuthorDB(db).get_papers_(followee, _one_year_back())

    # sort the papers
    sorted_authors = sorted(authors, key=lambda author: author.other_names)
    slice_ = slice(0, num_papers or None)
    papers: list[Paper] = sorted(
        set(chain.from_iterable([author.papers for author in sorted_authors])),
        key=lambda p: p.updated,
        reverse=True,
    )[slice_]

    def controller(stdscr: curses._CursesWindow, view: callable, papers: list) -> None:
        """Uses `curses` to listen for input, retrieves the relevant data and
        displays it using `view`.
        """
        with Storage() as db:
            cursor, key = 0, 0
            while key != ord("q"):
                paper = papers[cursor]
                view(paper, f"Triaged {cursor}/{len(papers)} papers")

                # wait for next input
                key = stdscr.getch()

                # controls
                match key:
                    case curses.KEY_DOWN:
                        cursor = max(0, cursor - 1)
                    case curses.KEY_LEFT:
                        cursor = min(len(papers) - 1, cursor + 1)
                    case curses.KEY_RIGHT:
                        cursor = min(len(papers) - 1, cursor + 1)

                # triage paper
                if key == curses.KEY_LEFT and paper.visible:
                    paper.visible = False
                    PaperDB(db).update(paper)
                    db.commit()
                elif key == curses.KEY_RIGHT and not paper.visible:
                    paper.visible = True
                    PaperDB(db).update(paper)
                    db.commit()

    def run(stdscr):
        """A thin callable for curses.wrapper()."""
        view = V.init_swipe_view(stdscr)
        controller(stdscr, view, papers)

    curses.wrapper(run)


def update(opt) -> None:
    if opt.author:
        with Storage() as db:
            author = AuthorDB(db).get(Author.from_string(opt.author).id)
            assert author.followed, "Can't update papers for authors you don't follow."
            followees = {author}
    else:
        with Storage() as db:
            followees = AuthorDB(db).get_followees()
        V.info(f"Updating {len(followees):n} authors you follow.")

    new_cnt, old_cnt, upd_cnt = 0, 0, 0
    db = Storage()
    for author in followees:
        try:
            AuthorArxiv.get_papers_(author)
        except UnexpectedEmptyPageError:
            V.info(f"error fetching papers by {author}.")
            continue

        if author.papers is None:
            log.info("update: no papers found for %s", author)
            continue

        print(f" {author:24s}  |  {len(author.papers):3d} papers.", end="\r")
        for paper in author.papers:
            if PaperDB(db).exists(paper):
                old = PaperDB(db).get(paper.id)
                if old.meta.version == paper.meta.version:
                    old_cnt += 1
                elif old.meta.version < paper.meta.version:
                    PaperDB(db).update(paper)
                    upd_cnt += 1
                else:
                    log.error("paper version can either be higher or the same.")
            else:
                # add paper
                PaperDB(db).save(paper)
                # and authors
                for author in paper.meta.authors:
                    if not AuthorDB(db).exists(author):
                        author._followed = False
                        AuthorDB(db).save(author)

                    # add the relationship.
                    # TODO: where should this one stay :)
                    db(
                        Q.add_authorship,
                        {"author_id": author.id, "paper_id": paper.id},
                    )
                new_cnt += 1
            # write to database
            db.commit()
    db.close()

    print("\x1b[2K", end="\r")  # clean line
    V.info("{:n} old, {:,} new, {:,} updated.".format(old_cnt, new_cnt, upd_cnt))


def follow(opt) -> None:
    log.debug(opt)
    result: list = list(AuthorArxiv.get(" ".join(opt.author)))
    with Storage() as db:
        match result:
            case []:
                V.info(f"{opt.author} not found on arXiv, try variations of the name.")
            case [author]:
                # TODO: checking status like this is ugly, there must be a better way
                if AuthorDB(db).exists(author):
                    if AuthorDB(db).is_followed_(author).followed:
                        V.info(f"{author} already followed.")
                        return
                    else:
                        author._followed = True
                        AuthorDB(db).update(author)
                else:
                    author._followed = True
                    AuthorDB(db).save(author)
                db.commit()

                followed = sorted(AuthorDB(db).get_followees())
                V.author_table(followed)
                V.info(f"\n{author} followed.")
                print(f"Following {len(followed)} authors.")
            case [*authors]:
                V.author_list(authors)
                V.info(f"\nMultiple matches for {opt.author}.")
                V.info("Try again with one of the authors above.")
            case _:
                log.error(f"pattern matching failed for {opt.author} -> {result}.")


def unfollow(opt) -> None:
    log.debug(opt)
    with Storage() as db:
        result: list = list(AuthorDB(db).search(" ".join(opt.author)))
        match result:
            case []:
                V.info(f"{opt.author} not in the database.")
            case [author]:
                # TODO: checking status like this is ugly, there must be a better way
                if not AuthorDB(db).is_followed_(author).followed:
                    V.info(f"{author} not followed.")
                    return
                author._followed = False
                AuthorDB(db).update(author)
                db.commit()
                followed = AuthorDB(db).get_followees()
                V.author_table(list(followed))
                V.info(f"\n{author} unfollowed.")
                V.info(f"Now following {len(followed)} authors.")
            case [*authors]:
                followed = AuthorDB(db).get_followees()
                hits_in_followed = sorted([a for a in authors if a in followed])
                V.info(f"Multiple matches for {opt.author}, out of which you follow:")
                V.author_list(hits_in_followed)
                V.info("Be more specific by using the full name.")
            case _:
                # TODO should we show the user non-critical errors?
                log.error(f"pattern matching failed for {opt.author} -> {result}.")


def info(opt: Info) -> None:
    # TODO: make a proper info view
    # TODO: fix when database has not authors or no papers

    cnts = {}
    with Storage() as db:
        cnts = {
            "papers": PaperDB(db).count(),
            "authors": AuthorDB(db).count(),
            "followed": AuthorDB(db).count_followees(),
        }
        followees = sorted(AuthorDB(db).get_followees())
        for followee in followees:
            AuthorDB(db).get_papers_(followee, "01-01-2020")

    print("data: {}\nlogs: {}".format(VOY_PATH, VOY_LOGS))

    # followees list
    if not followees:
        V.info("\nStart following authors with `voy add`.")
        return

    print("\nFollowing: ")
    V.author_table(followees)

    # other details
    print("\nDB details: ")
    for k, v in cnts.items():
        print(f"{k:16}{v:,}")
    with Storage() as db:
        last = PaperDB(db).last()
    print(
        "\nLast paper:\n[{}] {}".format(
            last.updated.split(" ")[0],
            last.meta.title,
        )
    )


def export(opt) -> None:
    with Storage() as db:
        followed = AuthorDB(db).get_followees()
    authors = sorted(followed, key=lambda a: (a.other_names, a.last_name))
    with open(opt.path, mode="w") as f:
        writer = csv.writer(f, delimiter=",", quotechar='"')
        for author in authors:
            writer.writerow([author.id, *author.names])


# argument parser config starts here  --------------------------------------------------


def _set_author_arg(default=False):
    return arg(
        default=() if default else MISSING,
        positional=True,
        help="eg.: Hinton | Geoff Hinton | Geoffrey Hinton",
    )


_pp = {"formatter_class": argparse.ArgumentDefaultsHelpFormatter}


@argsclass(description="list papers by authors you follow")
class Show:
    author: Sequence[str] = _set_author_arg(default=True)
    by_author: bool = arg(
        "-a",
        default=False,
        help="authors you follow and their papers (default: %(default)s)",
    )
    num: int = arg(
        "-n",
        default=0,
        help="no. of papers to list (default: 10); if using `--by_author` (default: 3)",
    )
    coauthors: bool = arg(
        "-c", default=False, help="show co-authors (default: %(default)s)"
    )
    url: bool = arg("-u", default=False, help="show arixv URL (default: %(default)s)")
    since: str = arg(
        "-t",
        default=(),
        help="earliest date to retrieve papers, eg. 01.01.2019 (default: one year ago)",
    )

    def __post_init__(self):
        self.num = self.num or (3 if self.by_author else 10)
        self.since = self.since or _one_year_back()

    def run(self) -> None:
        show(self)


@argsclass(
    description="""Opens a 'swipe' view that allows quickly going through the
    most recent papers and reviewing them based on abstract.

    Left arrow key flags the papers so that it does not show up again until next
    update and increments the list. Right arrow key increments the list.
    """
)
class Triage:
    def run(self) -> None:
        triage(self)


@argsclass(
    description="""Fetch papers for followed authors; by default uses the arxiv
    API.
    
        1) arxiv API author match (default): A query with the name of the author
        is made and a fuzzy match of the past 100 papers or so is returned.
        2) arxiv json bulk (wip): uses the kaggle json dump.
        3) arxiv API bulk (todo): fetch batches of K latest papers going back
        and update only the the authors you follow. 
    """,
    parser_params=_pp,
)
class Update:
    from_arxiv_json: Path | None = arg(
        help="path to the json downloaded from "
        + "`www.kaggle.com/datasets/Cornell-University/arxiv`",
    )
    author: Sequence[str] = _set_author_arg(default=True)
    # start_index: int = arg(default=1, help="starting arXiv API index")
    # stop_index: int = arg(default=10_001, help="maximum arXiv API index")

    def run(self) -> None:
        if self.from_arxiv_json:
            from_json(self)
        elif self.author:
            V.info("Updating a single author is not yet implemented.")
        else:
            update(self)


@argsclass(
    description="search authors or papers on arxiv or in the local database.",
    parser_params=_pp,
)
class Search:
    author: Sequence[str] = _set_author_arg(True)
    paper: str | None = arg(
        default=(),
        help="""eg.: arxiv_id | average-reward | attention is all you need.
        Search through titles and abstracts for the input terms.""",
    )
    db: bool = arg(
        "-db",
        default=False,
        help="""Search in the local database (default: %(default)s)""",
    )
    coauthors: bool = arg(
        "-c", default=False, help="show co-authors (default: %(default)s)"
    )
    url: bool = arg("-u", default=False, help="show arixv URL (default: %(default)s)")
    num: int = arg(
        "-n",
        default=2,
        help="""number of papers to show for each author when searching on arXiv
        (default: %(default)s)""",
    )

    def run(self):
        assert self.author or self.paper, (
            "Either search for authors or you search for papers."
        )

        if self.author and not self.db:
            search_author_in_arxiv(self.author, self.num)
        elif self.author and self.db:
            search_author_in_db(self.author)
        elif self.paper and not self.db:
            V.info("Paper search on arxiv is not yet implemented.")
        elif self.paper and self.db:
            search_text_in_db(self.paper, self.coauthors, self.url)
        else:
            log.debug("Some value error in voy.search: %s", self)
            V.info("Something fishy happened, check logs.")


@argsclass(description="follow an author", parser_params=_pp)
class Follow:
    author: Sequence[str] = _set_author_arg()

    def run(self):
        follow(self)


@argsclass(description="follow an author", parser_params=_pp)
class Unfollow:
    author: Sequence[str] = _set_author_arg()

    def run(self):
        unfollow(self)


@argsclass
class Info:
    pass

    def run(self):
        info(self)


@argsclass
class Export:
    path: Path = arg(default=Path.cwd() / "voy.csv", positional=True, help="some path")

    def run(self):
        export(self)


@argsclass
class Voy:
    action: Search | Show | Triage | Follow | Unfollow | Update | Info | Export

    def run(self):
        self.action.run()


def main() -> None:
    args = parse(
        Voy,
        parser=argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="Voy: a CLI for following arXiv authors.",
            epilog="""The flow would go like this:
    `voy search Remi Munos` to search for an author. You should see a list of
        authors and their papers, because authors often use slight variations of
        their name and because the arxiv API is fairly fuzzy.

    `voy follow Rémi Munos`         to follow one of the results from above.
    `voy follow Remi Munos`         to follow yet another name variation of the same author.
    `voy update --from-arxiv-api`   to fetch their papers and commit them to database.

    With the database updated you can now view the latest papers:

    `voy show`

    `voy show --help` to check all the options for displaying the papers.

    IMPORTANT: often the arxiv API returns no results, without error codes, even
    if the query is correct. Just let it cool for a while, it should be back eventualy.
            """,
        ),
    )
    log.info(args)

    args.run()


if __name__ == "__main__":
    main()



import logging
import time
from datetime import datetime as dt

import jsonlines as jl

from . import CATEGORIES
from . import query as Q
from . import views as V
from .models import Author, AuthorDB, Paper, PaperDB, PaperMeta
from .storage import Storage

BATCH_SIZE = 100

log = logging.getLogger("voy")


def from_json(opt) -> None:
    """Uses Kaggle dataset."""

    def arxiv2sqlite_datetime(ts):
        to = dt.strptime(ts, "%a, %d %b %Y %H:%M:%S %Z")
        return dt.strftime(to, Paper.DATE_FMT)

    t0 = time.time()
    db = Storage()
    with jl.open(opt.from_arxiv_json, "r") as f:
        for jid, _paper in enumerate(f):
            paper_categories = _paper["categories"].split(" ")
            if any([x in CATEGORIES for x in paper_categories]):
                # make the paper
                paper = Paper(
                    id=_paper["id"],
                    created=arxiv2sqlite_datetime(_paper["versions"][0]["created"]),
                    updated=arxiv2sqlite_datetime(_paper["versions"][-1]["created"]),
                    meta=PaperMeta(
                        version=int(_paper["versions"][-1]["version"][-1]),
                        authors=[Author(*a[:3]) for a in _paper["authors_parsed"]],
                        title=_paper["title"].replace("\n", "").replace("  ", " "),
                        abstract=_paper["abstract"]
                        .replace("\n", "")
                        .replace("  ", " "),
                        categories=paper_categories,
                    ),
                )
                # add the paper
                PaperDB(db).save(paper)

                # add all the authors
                for author in paper.meta.authors:
                    if not AuthorDB(db).exists(author):
                        AuthorDB(db).save(author)

                    # add the relationship
                    db(Q.add_authorship, {"author_id": author.id, "paper_id": paper.id})

                # write to database
                db.commit()

            if jid % 100_000 == 0 and jid != 0:
                delta = time.time() - t0
                paper_num, author_num = PaperDB(db).count(), AuthorDB(db).count()
                V.info(
                    "[{:7d}] papers={:6d}, authors={:6d}  |  {:6.1f} i/s.".format(
                        jid, paper_num, author_num, paper_num / delta
                    )
                )
    V.info(f"Papers:  {PaperDB(db).count():n}")
    V.info(f"Authors: , {AuthorDB(db).count():n}")
    db.close()



from __future__ import annotations

import json
import logging
import re
import time
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import asdict, dataclass
from datetime import datetime as dt

import arxiv
from xxhash import xxh3_64_hexdigest, xxh3_64_intdigest

from . import CATEGORIES
from . import query as Q
from .storage import Storage

log = logging.getLogger("voy")

PREFIX_MATCH = "van|der|de|la|von|del|della|da|mac|ter|dem|di|vaziri"
CAT_TERMS = " OR ".join([f"cat:{c}" for c in CATEGORIES])


__all__ = (
    "Author",
    "AuthorDB",
    "AuthorArxiv",
    "Paper",
    "PaperDB",
    "PaperArxiv",
    "PaperList",
)

PaperList = list["Paper"]


@dataclass
class Author:
    last_name: str
    other_names: str
    name_suffix: str | None = ""  # Why not = None?
    id: str | None = None
    _followed: bool | None = None
    _papers: PaperList | None = None

    def __post_init__(self):
        assert self.last_name.strip(), f"Author should have a last name, got {self}"
        # assert self.other_names.strip(), f"Should have a first name, got {self}."
        _id = xxh3_64_hexdigest(self.__hstr())  # f"{self.__hash__():x}"
        if self.id is None:
            self.id = _id
        assert self.id == _id, f"id mismatch: self._id={self.id}, computed={_id}"

    @classmethod
    def from_string(cls, string: str) -> Author:
        last, other, suffix = cls.normalize_author_name(string)
        return cls(last, other, suffix)

    @property
    def followed(self):
        return self._followed

    @property
    def papers(self):
        return self._papers

    @staticmethod
    def normalize_author_name(name: str) -> tuple[str, str, str]:
        """Copyright 2017 Cornell University

        Permission is hereby granted, free of charge, to any person obtaining a copy of
        this software and associated documentation files (the "Software"), to deal in
        the Software without restriction, including without limitation the rights to
        use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
        of the Software, and to permit persons to whom the Software is furnished to do
        so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.

        Credits: github.com/mattbierbaum/arxiv-public-datasets/
        """
        patterns = [
            (
                "double-prefix",
                r"^(.*)\s+(" + PREFIX_MATCH + r")\s(" + PREFIX_MATCH + r")\s(\S+)$",
            ),
            ("name-prefix-name", r"^(.*)\s+(" + PREFIX_MATCH + r")\s(\S+)$"),
            ("name-name-prefix", r"^(.*)\s+(\S+)\s(I|II|III|IV|V|Sr|Jr|Sr\.|Jr\.)$"),
            ("name-name", r"^(.*)\s+(\S+)$"),
        ]

        pattern_matches = (
            (mtype, re.match(m, name, flags=re.IGNORECASE)) for (mtype, m) in patterns
        )

        (mtype, match) = next(
            ((mtype, m) for (mtype, m) in pattern_matches if m is not None),
            ("default", None),
        )
        if match is None:
            author_entry = (name, "", "")
        elif mtype == "double-prefix":
            s = "{} {} {}".format(match.group(2), match.group(3), match.group(4))
            author_entry = (s, match.group(1), "")
        elif mtype == "name-prefix-name":
            s = "{} {}".format(match.group(2), match.group(3))
            author_entry = (s, match.group(1), "")
        elif mtype == "name-name-prefix":
            author_entry = (match.group(2), match.group(1), match.group(3))
        elif mtype == "name-name":
            author_entry = (match.group(2), match.group(1), "")
        else:
            author_entry = (name, "", "")

        return author_entry

    @property
    def names(self) -> tuple:
        return self.normalize_author_name(str(self))

    def dict(self):
        return {
            "last_name": self.last_name,
            "other_names": self.other_names,
            "name_suffix": self.name_suffix,
            "id": self.id,
            "followed": self._followed,
        }

    def __hstr(self):
        """Change this and the hash changes."""
        last, other, sfx = self.last_name, self.other_names, self.name_suffix
        return f"{other} {last} {sfx}" if sfx else f"{other} {last}"

    def __hash__(self):
        """Python's hash is randomized between runtimes.
        But we need a way to quickly index authors in the database.
        """
        return xxh3_64_intdigest(self.__hstr())

    def __eq__(self, other):
        if isinstance(other, Author):
            return str(self) == str(other)
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, Author):
            return self.last_name > other.last_name
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, Author):
            return self.last_name < other.last_name
        return NotImplemented

    def __str__(self):
        last, other, sfx = self.last_name, self.other_names, self.name_suffix
        return f"{other} {last} {sfx}" if sfx else f"{other} {last}"

    def __format__(self, format_spec):
        return format(str(self), format_spec)


@dataclass
class PaperMeta:
    title: str
    abstract: str
    authors: list[Author]
    version: int
    categories: list[str]


class Paper:
    DATE_FMT = "%Y-%m-%d %H:%M:%S"

    def __init__(self, id, created, updated, meta, visible) -> None:
        self.id: str = id
        self.created: str = created
        self.updated: str = updated
        self.meta: PaperMeta = meta
        self.visible: bool = visible
        self._post_init()

    def _post_init(self):
        # the id is not versioned (doesn't end in v1).
        assert self.id[-1] != "v", f"{self.id} is not a valid arxiv identifier."
        # date should work with sqlite
        try:
            dt.strptime(self.updated, Paper.DATE_FMT)
        except ValueError as err:
            msg = f"{self.updated} is not a valid sqlite datetime (%Y-%m-%d %H:%M:%S)"
            raise ValueError(f"{msg}, {err}")
        assert self.created <= self.updated, (
            f"Can't update before publishing:\n{repr(self)}."
        )

    def dict(self):
        d = {k: v for k, v in self.__dict__.items() if k[0] != "_"}
        return {**d, "meta": asdict(self.meta)}

    def __str__(self) -> str:
        return f"({self.id}/{self.updated}) {self.meta.title}"

    def __repr__(self) -> str:
        body = ", ".join([f"{k}={v}" for k, v in self.__dict__.items()])
        return f"Paper({body})"

    def __format__(self, format_spec):
        return format(str(self), format_spec)

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other):
        if isinstance(other, Paper):
            return self.id == other.id
        return NotImplemented


#  persistence layer


class Repository[T](ABC):
    @abstractmethod
    def get(self, id: str) -> T:
        raise NotImplementedError

    @abstractmethod
    def save(self, model: T) -> None:
        # TODO: should return something?
        raise NotImplementedError

    @abstractmethod
    def update(self, model: T) -> None:
        # TODO: should return something?
        raise NotImplementedError

    @abstractmethod
    def exists(self, model: T) -> bool:
        raise NotImplementedError

    @abstractmethod
    def count(self) -> int:
        raise NotImplementedError


class AuthorDB(Repository[Author]):
    """Implements SQLite Author persistence layer.

    It is meant to be used with a context manager:

    ```
    with Storage() as db:
        AuthorDB(db).get_papers(author)
    ```
    """

    def __init__(self, db: Storage) -> None:
        super().__init__()
        self.db = db

    def get(self, id: str) -> Author:
        res = self.db(Q.get_author, {"id": id}).fetchone()
        _id, last, other, suffix, followed = res
        return Author(last, other, suffix, _id, followed)

    # TODO: starting_date seems the wrong name
    def get_papers_(
        self,
        author: Author,
        starting_date: str,
        only_visible: bool = False,
    ) -> Author:
        """Retrieves the author's papers."""
        csr = self.db(
            Q.get_papers_by_author_id,
            {"aid": author.id, "starting_date": starting_date},
        )
        res = csr.fetchall()

        author._papers = [] if author._papers is None else author._papers
        for pid, updated, created, meta, visible in res:
            if only_visible and not visible:
                continue
            meta = PaperMeta(**json.loads(meta))
            author._papers.append(Paper(pid, created, updated, meta, bool(visible)))
        return author

    def get_followees(self) -> set[Author]:
        res = self.db(Q.get_followed, {}).fetchall()
        return set([Author(l, o, s, i, f) for i, l, o, s, f in res])

    def save(self, author: Author) -> None:
        self.db(Q.add_author, author.dict())

    def update(self, author: Author):
        """Only update that can happen is follow/unfollow."""
        self.db(Q.follow_author, {"id": author.id, "followed": int(author.followed)})

    def exists(self, author: Author) -> bool:
        csr = self.db(Q.is_author, {"id": author.id})
        return bool(csr.fetchone()[0])

    def is_followed_(self, author: Author) -> Author:
        if not self.exists(author):
            return author
        author._followed = bool(self.db(Q.is_followed, {"id": author.id}).fetchone()[0])
        return author

    def count(self) -> int:
        return self.db(Q.count_author, {}).fetchone()[0]

    def count_followees(self) -> int:
        return self.db(Q.count_followed, {}).fetchone()[0]

    def search(self, searched: str) -> set[Author]:
        last, other, suffix = Author.normalize_author_name(searched)
        log.debug("DB search %s, %s, %s", last, other, suffix)

        if other:
            csr = self.db(Q.search_author, {"last_name": last, "other_names": other})
            res = csr.fetchall()
        else:
            csr = self.db(Q.search_by_last_name, {"last_name": last})
            res = csr.fetchall()

        if not res:
            return set()
        # aid, last name, other names, suffix
        authors = set(sorted([Author(r[1], r[2], r[3], r[0], bool(r[4])) for r in res]))
        return authors


class PaperDB(Repository[Paper]):
    """Implements SQLite Paper persistence layer."""

    def __init__(self, db: Storage) -> None:
        super().__init__()
        self.db = db

    @staticmethod
    def _from_res(res):
        """Helper method that initializes a Paper."""
        id_, updated, created, meta, visible = res
        meta = PaperMeta(**json.loads(meta))
        return Paper(id_, created, updated, meta, bool(visible))

    def get(self, id: str) -> Paper:
        res = self.db(Q.get_paper, {"id": id}).fetchone()
        return self._from_res(res)

    def exists(self, paper: Paper) -> bool:
        csr = self.db(Q.is_paper, {"id": paper.id})
        return bool(csr.fetchone()[0])

    def save(self, paper: Paper) -> None:
        data = paper.dict()
        self.db(Q.add_paper, {**data, "meta": json.dumps(data["meta"])})

    def update(self, paper: Paper) -> None:
        data = paper.dict()
        self.db(Q.update_paper, {**data, "meta": json.dumps(data["meta"])})

    def count(self) -> int:
        return self.db(Q.count_paper, {}).fetchone()[0]

    def last(self, col: str = "created") -> Paper:
        res = self.db(Q.last_paper_by.format(col=col), {}).fetchone()
        return self._from_res(res)

    # TODO fix terms containing "-"
    def full_text_search(self, term: Sequence[str]) -> list[Paper]:
        res = self.db(Q.fts, {"term": term}).fetchall()
        return [self._from_res(r) for r in res]


class Rest[T](ABC):
    """A REST interface."""

    @abstractmethod
    def search(self, searched: str, max_results: int) -> set[T]:
        raise NotImplementedError


class AuthorArxiv(Rest[Author]):
    """Implements arXiv Author rest layer."""

    @staticmethod
    def _sanitize(q: str) -> str:
        return q.replace("'", "\\")  # "d'Oro -> d\Oro"

    @staticmethod
    def _make_query(last: str, other: str, suffix: str, with_and=False) -> str:
        """Not sure which is better:

        'au: last other':
            - seems to be the least selective, for common last names it will
            return a lot of them.
            - fails to return Pablo S. Castro for Pablo Samuel Castro or any of
            the variations.

        'au: last AND other':
            - fails to return Rémi Munos for Rémi Munos
        """
        sep = " AND " if with_and else " "
        sterm = sep.join([n for n in (last, other, suffix) if n])
        sterm = AuthorArxiv._sanitize(sterm)
        query = f"({CAT_TERMS}) AND (au:{sterm})"
        log.debug("arxiv search term: %s", sterm)
        return query

    @staticmethod
    def get(
        searched: str, max_results=100, order_by=arxiv.SortCriterion.Relevance
    ) -> set[Author]:
        last, other, suffix = Author.normalize_author_name(searched)

        # cast a wider net by ignoring the first other name
        _other = other.split(" ")[0] if " " in other else other
        matches: set[Author] = AuthorArxiv.search(" ".join([_other, last]), max_results)

        # exact match with the original search term
        for author in matches:
            if (other == author.other_names) and (suffix == author.name_suffix):
                return {author}
        return matches

    @staticmethod
    def search(
        searched: str, max_results=100, order_by=arxiv.SortCriterion.Relevance
    ) -> set[Author]:
        """Search the arXiv api for matching authors."""
        last, other, suffix = Author.normalize_author_name(searched)

        res = arxiv.Client().results(
            arxiv.Search(
                AuthorArxiv._make_query(last, other, suffix),
                sort_by=order_by,
                max_results=max_results,
            )
        )
        # extract the set of authors that match
        matches: dict[Author, PaperList] = defaultdict(list)
        for _paper in res:
            for _author in _paper.authors:
                author = Author.from_string(_author.name)
                if last in str(author):
                    matches[author].append(PaperArxiv._cast(_paper))
        for author, papers in matches.items():
            author._papers = papers
        log.info(f"AuthorArxiv search: got {len(matches):n} matches.")
        return set(matches.keys())

    @staticmethod
    def get_papers_(author: Author) -> Author:
        # TODO: one does not simply modify a complex type
        res = AuthorArxiv.get(str(author), order_by=arxiv.SortCriterion.LastUpdatedDate)
        if not res:
            res = AuthorArxiv.get(str(author), order_by=arxiv.SortCriterion.Relevance)
        if res:
            _author = res.pop()
            author._papers = _author.papers
            log.info(
                f"AuthorArxiv.papers: got {len(author.papers):n} papers for {author:s}."
            )
        return author


class PaperArxiv(Rest[Paper]):
    """Implements arXiv Author rest layer."""

    def search(self, searched: str, max_results: int) -> set[Paper]:
        raise NotImplementedError

    @staticmethod
    def _cast(paper: arxiv.Result) -> Paper:
        pid, version = paper.entry_id.split("/")[-1].split("v")

        return Paper(
            id=pid,
            created=time.strftime(Paper.DATE_FMT, paper._raw.published_parsed),
            updated=time.strftime(Paper.DATE_FMT, paper._raw.updated_parsed),
            meta=PaperMeta(
                title=paper.title,
                abstract=paper.summary,
                authors=[
                    Author(*Author.normalize_author_name(a.name)) for a in paper.authors
                ],
                version=version,
                categories=paper.categories,
            ),
            visible=True,
        )
